import numpy as np
import pandas as pd
from numba import jit, f8, i8


class Constrainter:

    def __init__(self, config):

        self.config = config

        self.discrete_constraints = self.config.discrete_constraints

        self.onehot_groups = self.config.onehot_groups

        self.onehot_constraints = self.config.onehot_constraints

        self.sumtotal_groups = self.config.sumtotal_groups

        self.sumtotal_constraints = self.config.sumtotal_constraints

        self.user_constraint_func = None

    def constraint(self, population):
        if not isinstance(population, np.ndarray):
            raise TypeError("populaiton must be ndarray")
        else:
            population = population.astype(np.float64)

        if self.onehot_constraints is not None:
            population = self.add_onehot_constraint(population)
        if self.sumtotal_constraints is not None:
            population = self.add_sumtotal_constraint(population)
        if self.discrete_constraints is not None:
            population = self.add_discrete_constraint(population)
        if self.user_constraint_func is not None:
            population = pd.DataFrame(population,
                                      columns=self.config.feature_names)
            population = self.user_constraint_func(population)
            population = population.values

        return population

    def add_discrete_constraint(self, population):
        for fname, constraints in self.discrete_constraints.items():
            idx = self.config.fname_to_idx(fname)
            constraints = np.array(constraints).astype(np.float64)
            population[:, idx] = _discrete(population[:, idx], constraints)

        return population

    def add_onehot_constraint(self, population):
        for uid, group in self.onehot_groups.items():
            columns = [self.config.fname_to_idx(fname) for fname in group]
            constraints = np.array(
                self.onehot_constraints[uid]).astype(np.float64)
            population[:, columns] = _onehot(population[:, columns],
                                             constraints)

        return population

    def add_sumtotal_constraint(self, population):
        for uid, group in self.sumtotal_groups.items():
            columns = [self.config.fname_to_idx(fname) for fname in group]
            constraints = np.array(
                self.sumtotal_constraints[uid]).astype(np.float64)
            population[:, columns] = _sumtotal(population[:, columns],
                                               constraints)
        return population


@jit(f8[:](f8[:], f8[:]), nopython=True)
def _discrete(arr, constraints):
    for i in range(arr.shape[0]):
        idx = (np.abs(constraints - arr[i])).argmin()
        arr[i] = constraints[idx]

    return arr


@jit(f8[:, :](f8[:, :], f8[:]), nopython=True)
def _onehot(arr, valuerange):
    #: if valuerange is [1, 1], equals to np.ones(arr.shape[0])
    lowerlim = valuerange[0]
    upperlim = valuerange[1]
    constants = np.random.uniform(lowerlim,
                                  upperlim,
                                  arr.shape[0])
    #: fill all zero rows (invalid rows) by constant
    arr[arr.sum(1) == 0] = 1

    #: create onehot array
    onehot_arr = np.zeros(arr.shape)
    nonzero_elements = (arr != 0)
    columns = np.arange(arr.shape[1])
    for i in range(arr.shape[0]):
        selected_col = np.random.choice(columns[nonzero_elements[i]])
        selected_val = arr[i, selected_col]
        if (selected_val <= upperlim) and (selected_val >= lowerlim):
            onehot_arr[i, selected_col] = selected_val
        else:
            onehot_arr[i, selected_col] = constants[i]

    return onehot_arr


@jit(f8[:, :](f8[:, :], f8[:]), nopython=True)
def _sumtotal(arr, valuerange):
    #: test onehot constraints
    lowerlim = valuerange[0]
    upperlim = valuerange[1]

    sumtotals = arr.sum(1)
    for i in range(arr.shape[0]):
        sumtotal = sumtotals[i]
        if lowerlim <= sumtotal and upperlim >= sumtotal:
            continue
        else:
            total_alt = valuerange[(np.abs(valuerange - sumtotal)).argmin()]
            arr[i] = (arr[i] / sumtotal) * total_alt

    return arr



import numpy as np
import pandas as pd


def get_onemax_samples(size, length):
    df = pd.DataFrame(np.random.rand(size, length))
    df.columns = [str(i) for i in range(length)]
    return df


def get_onemax_model():
    class model:
        def predict(self, samples):
            if not isinstance(samples, pd.DataFrame):
                raise TypeError("Samples must be DataFrame")
            return samples.sum(1)

    return model()


def get_linear_model(length):
    class model:
        def __init__(self, length):
            self.weights = np.random.randn(length).T

        def predict(self, samples):
            if not isinstance(samples, pd.DataFrame):
                raise TypeError("Samples must be DataFrame")
            return np.dot(samples, self.weights)

        def get_max_value(self):
            return self.weights[self.weights > 0].sum()

        def get_min_value(self):
            return self.weights[self.weights <= 0].sum()

    return model(length)



import copy
import os
import shutil
import uuid
from collections import defaultdict
from logging import DEBUG, Formatter, StreamHandler, getLogger

import numpy as np
import pandas as pd
from optga.config import OptConfig
from optga.evaluater import Evaluator
from optga.spawner import Spawner
from optga.strategy import EvolutionStrategy
from optga.tools.nsga2 import get_paretofront


def get_logger():
    logger = getLogger("RUN_GA")
    logger.setLevel(DEBUG)
    stream_handler = StreamHandler()
    handler_format = Formatter(
        '(%(levelname)s)[%(asctime)s]\n%(message)s')
    stream_handler.setFormatter(handler_format)
    logger.addHandler(stream_handler)
    return logger


class Optimizer:

    def __init__(self, samples):

        if not isinstance(samples, pd.DataFrame):
            raise Exception("Sample data must be pd.DataFrame")

        self.samples = samples

        self.config = OptConfig()

        self.single_constraints = {}

        self.group_constraints = {}

        self.user_constraint_func = None

        self.config.limits = self._get_limits(self.samples)

        self.models = {}

        self.history = {}

        self.pareto_front = {}

    def add_objective(self, objname, model, direction):
        #: validate model using samples
        try:
            model.predict(self.samples)
        except:
            raise Exception("Invalid samples, Try if model.predict(samples) works")

        if direction not in ["maximize", "minimize"]:
            raise KeyError(f'direction must be "maximize" or "minimize"')

        self.models[objname] = model

        if not self.config.objectives:
            self.config.objectives = {objname: direction}
        else:
            self.config.objectives[objname] = direction

    def add_valuerange_constraint(self, fname, lower, upper):
        try:
            self.config.limits[fname] = [lower, upper]
        except KeyError:
            print(f"{fname} not found from {list(self.config.limits.keys())}")

    def add_discrete_constraint(self, fname, constraints):
        """Add discrete_constraint

        Parameters
        ----------
        fname: str
            feature_name

        constraints : List[numeric]
            List of discrete values

        Raises
        ------
        Exception
            if fname no exists in self.config.feature_names
        """
        if fname not in self.config.feature_names:
            raise Exception(f"{fname} not in {self.config.feature_names}")

        if self.config.discrete_constraints is None:
            self.config.discrete_constraints = {fname: constraints}
        else:
            self.config.discrete_constraints[fname] = constraints

    def add_onehot_groupconstraint(self, group, lower=1.0, upper=1.0):
        for fname in group:
            if fname not in self.config.feature_names:
                raise Exception(f"{fname} not in {self.config.feature_names}")

        if lower > upper:
            lower, upper = upper, lower

        uid = str(uuid.uuid4())
        if self.config.onehot_constraints is None:
            self.config.onehot_groups = {uid: group}
            self.config.onehot_constraints = {uid: [lower, upper]}
        else:
            self.config.onehot_groups.update({uid: group})
            self.config.onehot_constraints.update({uid: [lower, upper]})

    def add_sumtotal_groupconstraint(self, group, lower, upper):
        """Add sum equal constraints
        if lower_lim == upper_lim, sum equal constraints

        Parameters
        ----------
        group : List[fname]
            List of feature_names
        lower : float
            lower lim
        upper : float
            upper lim

        """
        for fname in group:
            if fname not in self.config.feature_names:
                raise Exception(f"{fname} not in {self.config.feature_names}")

        if lower > upper:
            lower, upper = upper, lower

        uid = str(uuid.uuid4())
        if self.config.sumtotal_constraints is None:
            self.config.sumtotal_groups = {uid: group}
            self.config.sumtotal_constraints = {uid: [lower, upper]}
        else:
            self.config.sumtotal_groups.update({uid: group})
            self.config.sumtotal_constraints.update({uid: [lower, upper]})

    def add_user_constraint(self, func):
        self.user_constraint_func = func

    def spawn_population(self, n):
        self.compile()
        return pd.DataFrame(
            self.strategy.constraint(self.spawner.spawn(n).values),
            columns=self.config.feature_names)

    def export_config(self, config_path):
        with open(config_path, "w") as f:
            f.write(self.config.to_json(indent=2, ensure_ascii=False))

    def reload_config(self, config_path):
        with open(config_path, "r") as f:
            config_json = f.read()
        self.config = OptConfig.from_json(config_json)

    def show_config(self):
        config = self.config.to_json(indent=2, ensure_ascii=False)
        print(config)

    def export_result(self, export_dir):
        if os.path.exists(export_dir):
            print(f"Overwrite: {export_dir}")
            shutil.rmtree(export_dir)
        os.makedirs(export_dir)

        self.export_config(os.path.join(export_dir, "opt_config.json"))

        if "X_pareto" in self.pareto_front.keys():
            self.pareto_front["X_pareto"].to_csv(
                os.path.join(export_dir, "X_pareto.csv"), index=False)
            self.pareto_front["Y_pareto"].to_csv(
                os.path.join(export_dir, "Y_pareto.csv"), index=False)

            self.pareto_front["X_init"].to_csv(
                os.path.join(export_dir, "X_init.csv"), index=False)
            self.pareto_front["Y_init"].to_csv(
                os.path.join(export_dir, "Y_init.csv"), index=False)

        for obj_name in self.history.keys():
            df = self._get_history(self.history, obj_name)
            df.to_csv(os.path.join(export_dir, f"log_{obj_name}.csv"))

    def compile(self):

        self.spawner = Spawner(self.config)

        self.strategy = EvolutionStrategy(self.config)

        self.evaluator = Evaluator(self.config, self.models)

        if self.user_constraint_func is not None:
            self.strategy.constrainter.user_constraint_func = (
                self.user_constraint_func)

    def run(self, n_gen, population_size=500):
        """run evolutional optimization
        Todo: 最初と最後だけIndividualになる方式で

        Parameters
        ----------
        population_size : int
            number of individuals of one generation
        n_gen : int
            number of generation
        """
        logger = get_logger()
        logger.info(f"Settings:\n{self.show_config()}")

        history = defaultdict(lambda: {"Average": [],
                                       "MAX": [],
                                       "MIN": []})

        population = self.spawn_population(population_size).values
        self._validate()
        for n in range(n_gen):
            population, stats = self.run_generation(population,
                                                    population_size)

            for obj_name in self.config.objective_names:
                history[obj_name]["Average"].append(
                    stats.loc[obj_name, "Average"])
                history[obj_name]["MAX"].append(
                    stats.loc[obj_name, "MAX"])
                history[obj_name]["MIN"].append(
                    stats.loc[obj_name, "MIN"])

            if n % 10 == 0:
                logger.info(f"====Generation {n} ====")
                logger.info(stats)

        logger.info("GA optimization finished gracefully")
        for obj_name in self.config.objective_names:
            logger.info(f"====GA RESULT: {obj_name} ====")
            df = self._get_history(history, obj_name)
            self.history[obj_name] = df

        self._post_run(population)

    def run_generation(self, population, population_size):

        ancestor = copy.deepcopy(population)

        children = self.strategy.mate(population)
        children = self.strategy.mutate(children)
        children = self.strategy.constraint(children)

        offspring = pd.DataFrame(np.vstack([ancestor, children]),
                                 columns=self.config.feature_names)
        fitness = self.evaluator.evaluate(offspring)
        stats = self._get_stats(fitness)
        weighted_fitness = fitness * self.config.weights

        next_population = self.strategy.select(offspring.values,
                                               weighted_fitness,
                                               population_size)

        return next_population, stats

    def _get_limits(self, sample):
        limits = {}
        for col in sample.columns:
            limits[col] = [sample[col].min(),
                           sample[col].max()]
        return limits

    def _get_stats(self, fitness=None):
        stats = pd.DataFrame(index=self.config.objective_names)
        if fitness is None:
            return stats.T
        else:
            stats["MAX"] = fitness.max(0)
            stats["MIN"] = fitness.min(0)
            stats["Average"] = fitness.mean(0)
            return stats

    def _get_history(self, history, obj_name):
        df = pd.DataFrame()
        df["MAX"] = history[obj_name]["MAX"]
        df["MIN"] = history[obj_name]["MIN"]
        df["Average"] = history[obj_name]["Average"]
        return df

    def _post_run(self, population):
        population = pd.DataFrame(population,
                                  columns=self.config.feature_names)
        fitness = self.evaluator.evaluate(population)
        weighted_fitness = fitness * self.config.weights

        pareto_front = get_paretofront(population.values, weighted_fitness)
        pareto_front = pd.DataFrame(pareto_front,
                                    columns=self.config.feature_names)

        pareto_fitness = self.evaluator.evaluate(pareto_front)
        pareto_fitness = pd.DataFrame(pareto_fitness,
                                      columns=self.config.objective_names)

        self.pareto_front["X_pareto"] = pareto_front
        self.pareto_front["Y_pareto"] = pareto_fitness

        self.pareto_front["X_init"] = self.samples
        self.pareto_front["Y_init"] = pd.DataFrame(
            self.evaluator.evaluate(self.samples),
            columns=self.config.objective_names)

    def _validate(self):
        if not self.config.objectives:
            raise Exception("Objective not found")
        elif not len(self.models):
            raise Exception("Model not found")

        for key in self.config.objectives.keys():
            if key not in self.models:
                raise Exception(f"model {key} not found")

        if None in self.config.weights:
            raise Exception("Unexpected Error")


class PallarelOptimizer(Optimizer):

    def __init__(self, samples, n_jobs):
        super().__init__(samples)
        self.n_jobs = n_jobs



import numpy as np
import pandas as pd


class Spawner:
    """Spawner spawns population as pd.DataFrame
    """

    def __init__(self, config):
        self.config = config

        self.n_cols = len(self.config.feature_names)

    def spawn(self, n):
        population = np.zeros((n, self.n_cols))
        for i in range(self.n_cols):
            population[:, i] = np.random.uniform(self.config.lowerlim[i],
                                                 self.config.upperlim[i], n)
        population = pd.DataFrame(population,
                                  columns=self.config.feature_names)
        return population



from optga.tools.mate import MateCxTwoPoints
from optga.tools.mutate import MutateUniform
from optga.tools.select import SelectNSGA2
from optga.tools.constraint import Constrainter


class EvolutionStrategy:

    def __init__(self, config):
        self.config = config

        self.mater = globals()[self.config.mate](self.config)

        self.mutater = globals()[self.config.mutate](self.config)

        self.selecter = globals()[self.config.select]

        self.constrainter = Constrainter(self.config)

    def mate(self, population):

        offspring = self.mater.mate(population)

        return offspring

    def mutate(self, population):

        offspring = self.mutater.mutate(population)

        return offspring

    def select(self, population, fitness, size):

        selected_population = self.selecter(population, fitness, size)

        return selected_population

    def constraint(self, population):
        population = self.constrainter.constraint(population)
        return population



from dataclasses_json import dataclass_json
from dataclasses import dataclass
from typing import Dict, List

import numpy as np


@dataclass_json
@dataclass
class OptConfig:

    mate: str = "MateCxTwoPoints"

    mutate: str = "MutateUniform"

    select: str = "SelectNSGA2"

    birth_rate: int = 3

    mutpb: float = 0.2

    indpb: float = 0.1

    objectives: Dict[str, str] = None

    #: upperlim and lowerlim
    limits: Dict[str, List[float]] = None

    #: single constraints
    discrete_constraints: Dict[str, List[float]] = None

    #: onehot
    onehot_groups: Dict[str, List[str]] = None

    onehot_constraints: Dict[str, List[float]] = None

    #: sum equal, less, more group constraints
    sumtotal_groups: Dict[str, List[str]] = None

    sumtotal_constraints: Dict[str, List[float]] = None

    @property
    def weights(self):
        weights = [1.0 if val == "maximize" else -1.0 if "minimize" else None
                   for val in self.objectives.values()]
        return np.array(weights)

    @property
    def upperlim(self):
        return [val[1] for val in self.limits.values()]

    @property
    def lowerlim(self):
        return [val[0] for val in self.limits.values()]

    @property
    def objective_names(self):
        return [key for key in self.objectives.keys()]

    @property
    def feature_names(self):
        return [key for key in self.limits.keys()]

    @property
    def group_variables(self):
        groups = []
        for group in self.onehot_groups.values():
            groups.append(group)
        for group in self.sumtotal_groups.values():
            groups.append(group)
        return groups

    @property
    def group_variables_indices(self):
        groups = []
        for group in self.group_variables:
            groups.append([self.fname_to_idx(fname) for fname in group])
        return groups

    @property
    def indices_fnames_dict(self):
        return {i: fname for i, fname in enumerate(self.feature_names)}

    @property
    def fnames_indices_dict(self):
        return {fname: i for i, fname in enumerate(self.feature_names)}

    def fname_to_idx(self, fname):
        return self.fnames_indices_dict[fname]

    def idx_to_fname(self, idx):
        return self.indices_fnames_dict[idx]



from abc import abstractmethod

import numpy as np


class BaseMutate:

    def __init__(self, config):

        self.config = config

        self.upperlim = self.config.upperlim

        self.lowerlim = self.config.lowerlim

        self.mutpb = self.config.mutpb

        self.indpb = self.config.indpb

    @abstractmethod
    def mutate(self, population):
        raise NotImplementedError()


class MutateUniform(BaseMutate):

    def mutate(self, population):
        mask = [bool(np.random.binomial(1, self.mutpb))
                for i in range(population.shape[0])]

        population[mask] = self.mutate_(population[mask])
        return population

    def mutate_(self, population):
        mutation = np.zeros(population.shape)
        for i in range(mutation.shape[1]):
            mutation[:, i] = np.random.uniform(self.upperlim[i],
                                               self.lowerlim[i],
                                               mutation.shape[0])

        #: indpbで0のマスク
        mask = np.random.binomial(1, self.indpb, population.shape).astype(bool)
        mutation *= mask
        population *= ~mask

        return population + mutation



import numpy as np
from numba import jit, f8, i8


def NSGA2(population, fitness, k):
    fitness = fitness.astype(np.float64)
    dominate_table = get_dominatetable(fitness)
    fronts = get_fronts(dominate_table, k)
    sorted_fronts = [sort_by_CrowdingDist(fitness, front) for front in fronts]

    indices = []
    for front in sorted_fronts:
        indices += front
    next_population = population[indices[:k]]

    return next_population


def get_paretofront(population, fitness):
    fitness = fitness.astype(np.float64)
    dominate_table = get_dominatetable(fitness)
    front = get_fronts(dominate_table, k=1)
    return population[front[0]]


@jit(i8[:, :](f8[:, :]), nopython=True)
def get_dominatetable(fitness):
    n_fits = fitness.shape[0]

    dominate_table = np.zeros((n_fits, n_fits), dtype=np.int64)
    for i in range(n_fits):
        for j in range(i+1, n_fits):
            fit_i = fitness[i]
            fit_j = fitness[j]
            if np.all(fit_i == fit_j):
                continue
            elif np.all(fit_i >= fit_j):
                dominate_table[i, j] = 0
                dominate_table[j, i] = 1
            elif np.all(fit_j >= fit_i):
                dominate_table[i, j] = 1
                dominate_table[j, i] = 0

    return dominate_table


def get_fronts(dominate_table, k):
    #: (2) ランク順にソート
    K = min(k, dominate_table.shape[0])
    fronts = []
    selected_indices = []

    while len(selected_indices) < K:
        rank = dominate_table.sum(1)

        #: すでに選ばれているインデックスは除外する
        current_front = list(np.setdiff1d(np.where(rank == 0)[0],
                             selected_indices))
        fronts.append(current_front)
        selected_indices += current_front

        #: 支配の解除
        dominate_table[:, current_front] = 0

    return fronts


def sort_by_CrowdingDist(fitness, front):
    distances = np.zeros(len(front))

    fitness_ = fitness[front]
    sorted_indices = np.argsort(fitness_, axis=0)
    norms = fitness_.max(0) - fitness_.min(0)

    for i in range(1, sorted_indices.shape[0]):
        for j in range(sorted_indices.shape[1]):
            if i == 0:
                distances[sorted_indices[i, j]] = np.inf
            elif i == (sorted_indices.shape[0]-1):
                distances[sorted_indices[i, j]] = np.inf
            elif norms[j] == 0:
                continue
            else:
                prev_val = fitness_[sorted_indices[i-1, j], j]
                next_val = fitness_[sorted_indices[i+1, j], j]
                distances[sorted_indices[i, j]] += (next_val - prev_val) / norms[j]

    sorted_front = sorted(zip(front, distances), key=lambda x: x[1], reverse=True)
    sorted_front = [idx for idx, dist in sorted_front]
    return sorted_front



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['optga']

package_data = \
{'': ['*'], 'optga': ['tools/*']}

install_requires = \
['dataclasses_json>=0.3.7,<0.4.0',
 'numba>=0.46.0,<0.47.0',
 'numpy>=1.17,<2.0',
 'pandas>=0.25.3,<0.26.0']

entry_points = \
{'console_scripts': ['main = tests.test_optga:main']}

setup_kwargs = {
    'name': 'optga',
    'version': '0.1.1',
    'description': 'Framework to find out optimal input of machine learning model',
    'long_description': None,
    'author': 'horoiwa',
    'author_email': 'horoiwa195@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABCMeta, abstractmethod
import copy

import numpy as np
from numba import jit, f8, i8


class BaseMate:

    def __init__(self, config):
        self.config = config

        self.birth_rate = self.config.birth_rate

        self.groups = self.config.group_variables_indices

    @abstractmethod
    def mate(self, population):
        raise NotImplementedError()


class MateCxTwoPoints(BaseMate):

    def mate(self, population):
        population_clone = copy.deepcopy(population)
        population_clone = np.random.permutation(
            np.vstack([population_clone for _ in range(self.birth_rate)]))
        population = np.vstack([population for _ in range(self.birth_rate)])

        cxpoints = np.array([(min(a, b), max(a, b)) if a != b else (a-1, b+1)
                             for a, b in zip(
                                 np.random.randint(1, population.shape[1]-1,
                                                   population.shape[0]),
                                 np.random.randint(1, population.shape[1]-1,
                                                   population.shape[0]))])

        mask = apply_cxpoints(np.ones(population.shape).astype(np.int64),
                              cxpoints.astype(np.int64))

        #: group variables must be exchanged together
        for group in self.groups:
            group = np.array(group).astype(np.int64)
            mask[:, group] = adjust_group(mask[:, group])

        mask = mask.astype(bool)
        population *= mask
        population_clone *= ~mask

        children = population + population_clone
        return children


@jit(i8[:, :](i8[:, :], i8[:, :]), nopython=True)
def apply_cxpoints(mask, cxpoints):
    for i in range(mask.shape[0]):
        mask[i, cxpoints[i, 0]:cxpoints[i, 1]] = 0
    return mask


@jit(i8[:, :](i8[:, :]), nopython=True)
def adjust_group(arr, ):
    for i in range(arr.shape[0]):
        arr[i, :] = np.round(np.mean(arr[i, :]))
    return arr


if __name__ == "__main__":
    arr1 = np.arange(1, 13).reshape(3, 4)
    arr2 = np.zeros((3, 4))



import numpy as np


class Evaluator:

    def __init__(self, config, models):

        self.config = config

        self.objective_names = (list(self.config.objectives)
                                if self.config.objectives else [])

        self.models = models

    def evaluate(self, population):
        fitness = np.zeros((population.shape[0], len(self.objective_names)))
        for i, oname in enumerate(self.objective_names):
            fitness[:, i] = self.models[oname].predict(population)
        return fitness
